#pragma once

#if defined(__cplusplus)
extern "C" {
#endif

#include <lektor/common.h>
#include <lektor/uri.h>
#include <lektor/lib/strv.h>

#include <sqlite3.h>
#include <mpv/client.h>

// clang-format off

struct lkt_queue_state {
    int volume;
    int paused;
    int random;
    int repeat;
    int single;
    int current;
    int duration;
    int consume;
    int length;
};

struct kara_metadata;
struct kara_metadata_rules;

#define database_version(str, CAPS, i) LKT_DATABASE_VERSION_ ##CAPS = i,
typedef enum {
#include <lektor/internal/dbversion.def>
    ___LKT_DATABASE_VERSION_COUNT,
} LKT_DATABASE_VERSION;
#undef database_version

/* Always the latest version */
#define LKT_DATABASE_VERSION_LATEST ((LKT_DATABASE_VERSION)(___LKT_DATABASE_VERSION_COUNT - 1))

/* Update stuff */
void database_get_update  (lkt_db *db, long *timestamp, long *job, int *current);
void database_stamp       (lkt_db *db);
void database_updated     (lkt_db *db);
void database_deleted_kara(lkt_db *db, int **kara_id, size_t *len);
void database_total_playtime(lkt_db *db, uint64_t *seconds);
int  database_upgrade_scheme(va_list *db);
LKT_DATABASE_VERSION database_get_version(lkt_db *db);
const char *database_get_version_name(LKT_DATABASE_VERSION);

/* Open correctly a database for lektor. */
bool database_new   (lkt_db **db);
bool database_init  (const char *dbpath);
bool database_open  (lkt_db *db, const char *dbpath, bool check);
bool database_attach(lkt_db *db, const char *name, const char *dbpath);
bool database_detach(lkt_db *db, const char *name);
void database_close_all(void);

/* Get information on the queue and currently playing kara. */
bool database_queue_state       (lkt_db *db, struct lkt_queue_state *res);
bool database_queue_current_kara(lkt_db *db, struct kara_metadata *res, int *id);
void database_queue_playtime    (lkt_db *db, uint64_t *seconds);

bool database_queue_set_paused       (lkt_db *db, bool paused);
bool database_queue_set_current_index(lkt_db *db, int idx);
bool database_queue_get_current_file (lkt_db *db, char filepath[PATH_MAX]);

bool database_get_kara_id           (lkt_db *db, char filepath[PATH_MAX], int *id);
bool database_get_kara_path         (lkt_db *db, int id, char filepath[PATH_MAX]);
bool database_get_kara_path_no_cache(lkt_db *db, int id, char filepath[PATH_MAX]);
bool database_get_kara_position     (lkt_db *db, int id, int *pos);
bool database_get_kara_mtime_id     (lkt_db *db, int id, uint64_t *mtime);
bool database_get_kara_duration_id  (lkt_db *db, int id, uint64_t *duration);
bool database_get_kara_mtime_path   (lkt_db *db, char filepath[PATH_MAX], uint64_t *mtime);
bool database_get_kara_duration_path(lkt_db *db, char filepath[PATH_MAX], uint64_t *duration);
bool database_is_kara_mkv           (lkt_db *db, char filepath[PATH_MAX], bool *valid_mkv);
bool database_is_kara_like          (lkt_db *db, int id, const char *str);
bool database_is_kara_available     (lkt_db *db, int id, bool *avail);
void database_cache_kara            (lkt_db *db, int id);
void database_cache_kara_magic      (lkt_db *db, int id);

bool database_get_kara_repo_timestamp_id  (lkt_db *db, int id, uint64_t *timestamp);
bool database_get_kara_repo_timestamp_path(lkt_db *db, char[PATH_MAX], uint64_t *timestamp);
bool database_set_kara_repo_timestamp_id  (lkt_db *db, int id, uint64_t timestamp);

/* Get a list of all non-cached karas from the disk */
void database_get_all_non_cached_kara(lkt_db *db, int **ret_ids, size_t *ret_ids_len);

/* Get a list of all cached karas from the memory */
void database_get_all_cached_kara(lkt_db *db, int **ret_ids, size_t *ret_ids_len);

/* Update the database. */
bool database_update        (lkt_db *db, const char *kara_dir, void *maybe_worker, int check_timestamp);
bool database_update_add    (lkt_db *db, const char *kara_path, struct kara_metadata *mdt, uint64_t id, bool avail);
void database_update_set_new(lkt_db *db, int id, bool is_new);
void database_update_del    (lkt_db *db, int id);
void database_update_touch  (lkt_db *db, int id);
void database_update_set_unavailable(lkt_db *db, int id); /* Mark a kara unavailable, for when it is updated! */

/* Control the content of the queue. */
bool database_queue_add_uri(lkt_db *db, struct lkt_uri *uri, int priority);
bool database_queue_add_id (lkt_db *db, int id,              int priority);
bool database_queue_del_id (lkt_db *db, int id);
bool database_queue_del_pos(lkt_db *db, int pos);
bool database_queue_clear  (lkt_db *db);
bool database_queue_crop   (lkt_db *db);
bool database_queue_flat   (lkt_db *db);
bool database_queue_move   (lkt_db *db, int from, int to);
bool database_queue_swap   (lkt_db *db, int from, int to);
bool database_queue_shuffle(lkt_db *db);
bool database_queue_seekid (lkt_db *db, int id, int *out_pos);
bool database_queue_dump   (lkt_db *db, const char *plt_name);

/* Add multiple IDs at once to the queue. The `int *` is the pointy thing that
 * STB returns to you. Be warned that all the queries will be done 10 per 10,
 * so if you do an add with 100 items, it will do 10 requests.
 * The function returns `true` if the adds where successfull, `false` if any
 * error occured.
 * TODO: Add a way to manually set the step count. */
bool database_queue_multi_add_id(lkt_db *, size_t *ids_stb_list, int priority);

/* Control the playing state of the queue. */
bool database_queue_toggle_pause(lkt_db *db);
bool database_queue_play        (lkt_db *db, int pos);
bool database_queue_stop        (lkt_db *db);

/* A search callback to be called after each matched row */
struct lkt_callback {
    bool (*call)(void *args, int pos, int pos_len, int id, int id_len, const char *sql_row);
    struct lkt_state *srv;
    size_t c;
    int iterations;
};

/* List the content of the queue */
bool database_queue_list(lkt_db *db, size_t from, size_t to, struct lkt_callback *callback);
int database_queue_probe_id(lkt_db *db, int id);

/* Sticker construct structure */
struct lkt_sticker_opt {
    const char *name;   /* Sticker name                      */
    int value;          /* The value of a sticker            */
    int uri;            /* URI of a sticker                  */
    char op;            /* Comparaison operator for stickers */
    const char *type;   /* Type of sticker                   */
};
#define LKT_STICKER_OPT_INIT { .name = NULL, .value = 0, .uri = 0, .op = '\0', .type = NULL }

/* Search the database */
typedef bool (*lkt_search_init_add_func)    (lkt_db *, struct lkt_uri *, int);
typedef bool (*lkt_search_database_func)    (struct lkt_state *srv, size_t c, int id, int id_len, const char *row);
typedef bool (*lkt_search_queue_func)       (struct lkt_state *srv, size_t c, int id, int id_len, const char *row);
typedef bool (*lkt_search_sticker_func)     (struct lkt_state *srv, size_t c, const char *sticker, const char *type, int uri, int value);
typedef bool (*lkt_search_listplaylist_func)(struct lkt_state *srv, size_t c, const char *name);

struct lkt_search *database_search_new(struct lkt_state *srv, size_t c, int cont, function_ptr call);
struct lkt_search *database_search_new_from(struct lkt_search *);

void database_search_set_uri(struct lkt_search *, struct lkt_uri *);
void database_search_set_name(struct lkt_search *, const char *);
void database_search_set_continuation(struct lkt_search *, size_t);
void database_search_set_sticker(struct lkt_search *, struct lkt_sticker_opt *);
struct lkt_state* database_search_get_srv(struct lkt_search *);
size_t database_search_get_client(struct lkt_search *);
size_t database_search_get_continuation(struct lkt_search *);

/* Free a search. The database_search_iter will free the search upon exit. This
 * function is here if you need to abort a search. */
void database_search_free(struct lkt_search *);

typedef bool (*database_search_init_func)(lkt_db *, struct lkt_search *);
bool database_search_database_init    (lkt_db *db, struct lkt_search *ret);
bool database_search_queue_init       (lkt_db *db, struct lkt_search *ret);
bool database_search_sticker_init     (lkt_db *db, struct lkt_search *ret);
bool database_search_playlist_init    (lkt_db *db, struct lkt_search *ret);
bool database_search_listplaylist_init(lkt_db *db, struct lkt_search *ret);

/* Iterate through an initialized search structure, will free it when the
 * search is done (return `false`). */
bool database_search_iter(struct lkt_search *item);

bool database_kara_by_id(lkt_db *db, int id, struct kara_metadata *kara, char filepath[PATH_MAX]);

/* Next and prev operation on the queue. */
bool database_queue_next        (lkt_db *db, char filepath[PATH_MAX]);
bool database_queue_prev        (lkt_db *db, char filepath[PATH_MAX]);
bool database_queue_skip_current(lkt_db *db, char filepath[PATH_MAX]);

/* Set a value in the config table */
bool database_validate_conf             (lkt_db *db);
bool database_config_set                (lkt_db *db, const char *section, const char *key, const char *value);
bool database_config_set_strv_1_2       (lkt_db *db, const char *section, const struct strv, const struct strv);
bool database_config_get_text           (lkt_db *db, const char *section, const char *key, char *value, size_t len);
bool database_config_get_text_nospace   (lkt_db *db, const char *section, const char *key, char *value, size_t len);
bool database_config_get_int            (lkt_db *db, const char *section, const char *key, int *value);
bool database_config_get_octal          (lkt_db *db, const char *section, const char *key, unsigned int *value);
bool database_config_exists             (lkt_db *db, const char *section, const char *key);
bool database_config_queue              (lkt_db *db, const char *option,  int value);
bool database_config_queue_default      (lkt_db *db);
void database_config_set_optional_fields(lkt_db *db);

#define DBCONF_GET_TEXT(db, section, key, value, len) \
    FAIL_UNLESS(database_config_get_text(db, section, key, value, len), \
                "Failed to get config '%s/%s' from database", section, key);
#define DBCONF_GET_TEXT_NOSPACE(db, section, key, value, len) \
    FAIL_UNLESS(database_config_get_text_nospace(db, section, key, value, len), \
                "Failed to get config '%s/%s' from database", section, key);
#define DBCONF_GET_INT(db, section, key, value) \
    FAIL_UNLESS(database_config_get_int(db, section, key, value), \
                "Failed to get config '%s/%s' from database", section, key);
#define DBCONF_GET_OCTAL(db, section, key, value) \
    FAIL_UNLESS(database_config_get_octal(db, section, key, value), \
                "Failed to get config '%s/%s' from database", section, key);

/* Dump configuration to any stream, form is undefined, this function is here
 * for debug purpose. */
bool database_config_dump(lkt_db *db, FILE *f);

/* Check obfuscation status. If no one is set, set it and return `true`. If
 * there was already an obfuscation status, return `true` if the two are the
 * same, `false` otherwise. */
bool database_config_obfuscation_check(lkt_db *db, int obfuscation);

/* Playlists */
struct lkt_playlist_metadata {
    const char *name;
    int last_update;
    int content_count;
};

/* Control playlists */
bool database_plt_touch     (lkt_db *db, const char *name);
bool database_plt_create    (lkt_db *db, const char *name);
bool database_plt_remove    (lkt_db *db, const char *name);
bool database_plt_remove_pos(lkt_db *db, const char *name, int pos);
bool database_plt_clear     (lkt_db *db, const char *name);
bool database_plt_rename    (lkt_db *db, const char *old_name, const char *new_name);

bool database_plt_export(lkt_db *db, const char *name);
bool database_plt_import(lkt_db *db, const char *name);

bool database_plt_add_uri(lkt_db *db, const char *name, struct lkt_uri *uri);

bool database_plt_contais_kara(lkt_db *db, const char *name, int karaid);

/* User control, yeah, MPD authentification sucks. */
bool database_user_authentificate(lkt_db *db, const char *password);
bool database_user_add(lkt_db *db, const char *username, const char *password);

/* Stickers manipulations. */
bool database_sticker_create        (lkt_db *db, const char *name);
bool database_sticker_delete        (lkt_db *db, const char *name);
bool database_sticker_delete_specify(lkt_db *sb, const char *type, int uri, const char *name);
bool database_sticker_set           (lkt_db *db, const char *type, const char *name, int uri, int value);

/* Get some stats from the database */
bool database_stats(lkt_db *db, int *authors, int *sources, int *karas);

// clang-format on

#if defined(__cplusplus)
}
#endif
